import json
import logging
from collections.abc import Callable
from typing import Literal, TypeAlias

from ..model import EventMessage

logger = logging.getLogger("compose")

HookEventType = Literal[
    "on_start",
    "on_receive",
    "on_receive_error",
    "on_consume",
    "on_consume_error",
    "on_shutdown",
]
HookArgType: TypeAlias = str | EventMessage | Exception
Hook: TypeAlias = Callable[[HookArgType], None]


def default_hook(_) -> None: ...


def log_event_message(log_message: str, message: EventMessage) -> None:
    logger.info(
        f"{log_message}: {json.dumps(message.encode())}",
        extra={"event_message": message.encode()},
    )


def log_exception(log_message: str, exc: Exception) -> None:
    logger.exception(log_message, exc_info=exc, stack_info=True)


DEFAULT_HOOKS = {
    "on_start": [logger.info],
    "on_receive": [lambda message: log_event_message("Received message", message)],
    "on_receive_error": [lambda exc: log_exception("Failed to receive message", exc)],
    "on_consume": [lambda message: log_event_message("Consumed message", message)],
    "on_consume_error": [lambda exc: log_exception(f"Failed to consume message due to {exc}", exc)],
    "on_shutdown": [logger.info],
}



from collections.abc import Callable
from typing import Annotated, Any, TypeVar, get_args

from fastapi import Depends, Query
from pydantic import BaseModel, Field, Json, create_model

from compose import compat, container

if compat.IS_PYDANTIC_V2:
    import pydantic_core
    from pydantic import field_validator

Q = TypeVar("Q", bound=BaseModel)
T = TypeVar("T", bound=container.BaseModel)


def dict_to_json(v: dict[str, Any] | None) -> str | None:
    if v is None:
        return v

    return pydantic_core.to_json(v).decode()


if compat.IS_PYDANTIC_V2:
    TYPE_VALIDATORS = {
        Json: [
            dict_to_json,
        ]
    }


def to_query(q: type[Q], /) -> type[Q]:
    field_args = (
        "title",
        "alias",
        "default",
        "default_factory",
        "description",
    )

    if compat.IS_PYDANTIC_V2:
        validators = {}
        field_args = (*field_args, "annotation")
        field_definitions = {}
        for field_name, field_info in q.model_fields.items():
            annotation = field_info.annotation
            field_definitions[field_name] = (
                annotation,
                Field(Query(**{arg: getattr(field_info, arg, None) for arg in field_args})),
            )
            if not (args := get_args(annotation)):
                continue

            if (arg := next((arg for arg in args if arg is not None), None)) is None:
                continue

            validators |= {
                f"{field_name}_{validator.__name__}": field_validator(field_name, mode="before")(
                    validator
                )
                for validator in TYPE_VALIDATORS.get(arg, [])
            }

        return create_model(
            f"{q.__name__}Query",
            **field_definitions,
            __validators__=validators,
            __base__=q,
        )

    else:
        field_definitions = {
            field_name: (
                field.outer_type_,
                Field(Query(**{arg: getattr(field, arg, None) for arg in field_args})),
            )
            for field_name, field in q.__fields__.items()
        }

        class Child(q):
            class Config:
                arbitrary_types_allowed = True

        return create_model(
            f"{q.__name__}Query",
            **field_definitions,
            __base__=Child,
        )


def as_query(q: type[Q], /) -> type[Q]:
    return Annotated[q, Depends(to_query(q))]


def create_model_dependency_resolver(
    model_type: type[T],
    dependencies: dict[str, tuple[type, Any]],
) -> Callable[..., Any]:
    dependencies_model = create_model(
        f"{model_type.__name__}fields",
        **{
            name: (field_type, Field(field_value))
            for name, (field_type, field_value) in dependencies.items()
        },
    )

    def wrapper(
        t: model_type,
        resolved_dependencies: Annotated[dependencies_model, Depends(dependencies_model)],
    ) -> T:
        return t.copy(update=resolved_dependencies.dict(), deep=True)

    return wrapper


def with_depends(model_type: type[T], **params: Any) -> type[T]:
    return Annotated[
        model_type,
        Depends(create_model_dependency_resolver(model_type, params)),
    ]



import http
from collections.abc import Awaitable, Callable
from typing import ClassVar, Self, TypeAlias, TypeVar

from fastapi import Request, Response
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
from pydantic import ValidationError
from starlette.exceptions import HTTPException

from compose import compat, schema

E: TypeAlias = TypeVar("E", bound=Exception)
ExceptionHandler: TypeAlias = Callable[[Request, E], Response | Awaitable[Response]]


class ExceptionHandlerInfo:
    default_response_cls: ClassVar[type[Response]] = JSONResponse

    def __init__(
        self, exc_class_or_status_code: int | type[Exception], handler: ExceptionHandler
    ) -> None:
        self.exc_class_or_status_code = exc_class_or_status_code
        self.handler = handler

    @classmethod
    def for_status_code(
        cls,
        status_code: http.HTTPStatus,
        error_type: str | None = None,
        response_cls: type[Response] | None = None,
    ) -> Self:
        return cls(
            exc_class_or_status_code=status_code,
            handler=create_exception_handler(
                status_code=status_code,
                error_type=error_type or status_code.name.lower(),
                response_cls=response_cls or cls.default_response_cls,
            ),
        )

    @classmethod
    def for_exc(
        cls,
        exc_cls: type[Exception],
        status_code: int,
        error_type: str | None = None,
        response_cls: type[Response] | None = None,
    ) -> Self:
        return cls(
            exc_class_or_status_code=exc_cls,
            handler=create_exception_handler(
                status_code=status_code,
                error_type=error_type or http.HTTPStatus(status_code).name.lower(),
                response_cls=response_cls or cls.default_response_cls,
            ),
        )

    @classmethod
    def default(cls) -> Self:
        return cls(
            exc_class_or_status_code=Exception,
            handler=create_exception_handler(
                status_code=http.HTTPStatus.INTERNAL_SERVER_ERROR,
                error_type=http.HTTPStatus.INTERNAL_SERVER_ERROR.name.lower(),
                response_cls=cls.default_response_cls,
            ),
        )

    @classmethod
    def for_http_exception(cls, response_cls: type[Response] | None = None) -> Self:
        return cls(
            exc_class_or_status_code=HTTPException,
            handler=create_http_exception_handler(response_cls or cls.default_response_cls),
        )

    @classmethod
    def for_request_validation_error(
        cls,
        exception_handler: ExceptionHandler | None = None,
        response_cls: type[Response] | None = None,
    ) -> Self:
        return cls(
            exc_class_or_status_code=RequestValidationError,
            handler=(
                exception_handler
                or create_default_validation_error_handler(response_cls or cls.default_response_cls)
            ),
        )

    @classmethod
    def for_pydantic_validation_error(
        cls,
        exception_handler: ExceptionHandler | None = None,
        response_cls: type[Response] | None = None,
    ) -> Self:
        return cls(
            exc_class_or_status_code=ValidationError,
            handler=(
                exception_handler
                or create_default_validation_error_handler(response_cls or cls.default_response_cls)
            ),
        )


def create_exception_handler(
    status_code: int,
    error_type: str,
    response_cls: type[Response],
) -> ExceptionHandler:
    def exception_handler(request: Request, exc: Exception) -> Response:
        response = schema.Error(
            title=str(exc),
            type=error_type,
            detail=getattr(exc, "detail", None),
            invalid_params=(
                (invalid_params := getattr(exc, "invalid_params", None))
                and [
                    compat.model_validate(t=schema.InvalidParam, obj=invalid_param)
                    for invalid_param in invalid_params
                ]
            ),
        )
        return response_cls(content=jsonable_encoder(response), status_code=status_code)

    return exception_handler


def create_http_exception_handler(response_cls: type[Response]) -> ExceptionHandler:
    def http_exception_handler(request: Request, exc: HTTPException) -> Response:
        status_code = http.HTTPStatus(exc.status_code)
        return response_cls(
            content=jsonable_encoder(
                schema.Error(
                    title=exc.detail,
                    type=status_code.name.lower(),
                )
            ),
            status_code=status_code,
            headers=exc.headers,
        )

    return http_exception_handler


def create_default_validation_error_handler(response_cls: type[Response]) -> ExceptionHandler:
    def validation_error_handler(
        request: Request, exc: RequestValidationError | ValidationError
    ) -> Response:
        return response_cls(
            content=jsonable_encoder(
                schema.Error.from_validation_error(exc=exc, title="Validation failed"),
            ),
            status_code=http.HTTPStatus.UNPROCESSABLE_ENTITY,
        )

    return validation_error_handler



import functools
import warnings
from collections.abc import Callable
from pathlib import Path

import pytest

from compose.enums import AppEnv

from .enums import TestTypeMarker


@functools.lru_cache(1)
def get_marker_name_to_marker() -> dict[str, pytest.MarkDecorator]:
    return {
        marker_name: getattr(pytest.mark, marker_name) for marker_name in TestTypeMarker.names()
    }


def check_env_is_allowed(env: AppEnv) -> None:
    if env is None or env != AppEnv.TEST:
        raise RuntimeError(f"`APP_ENV` must be set to `{AppEnv.TEST}` to run tests")


def register_test_type_markers(config: pytest.Config) -> None:
    for member in TestTypeMarker:
        name, description = member.value
        config.addinivalue_line("markers", f"{name}: {description}")


def default_marker_getter(item: pytest.Function) -> pytest.MarkDecorator:
    with warnings.catch_warnings(
        action="ignore",
        category=pytest.PytestUnknownMarkWarning,
    ):
        marker_name_to_marker = get_marker_name_to_marker()
        test_types = set(marker_name_to_marker.keys())
        default_test_type = pytest.mark.unit.name

        node_path = item.nodeid.split("::")[0]
        parts = Path(node_path).parts

        mark_name = next((part for part in parts if part in test_types), default_test_type)
        result = getattr(pytest.mark, mark_name)

    return result


def add_test_type_markers(
    items: list[pytest.Function],
    marker_getter: Callable[[pytest.Function], pytest.MarkDecorator] = default_marker_getter,
) -> None:
    """

    Examples:
        >>> from compose import testing
        >>> import pytest
        >>> def pytest_collection_modifyitems(items: list[pytest.Function]) -> None:
        >>>     testing.add_test_type_markers(items)

    """

    for item in items:
        item.add_marker(marker_getter(item))



from typing import Self

from compose.types import Str

try:
    import bcrypt
except ImportError:
    raise ImportError("bcrypt is required for password hashing")


class HashedPassword(Str):
    @classmethod
    def hash(cls, password: str) -> Self:
        return cls(bcrypt.hashpw(password=password.encode(), salt=bcrypt.gensalt()).decode())

    def verify(self, password: str) -> bool:
        return bcrypt.checkpw(password=password.encode(), hashed_password=self.encode())



from __future__ import annotations

import datetime
from collections.abc import Callable, Generator
from typing import Any

import pendulum

from .. import compat
from .helper import get_pydantic_core_schema

if compat.IS_PYDANTIC_V2:
    from pydantic import GetCoreSchemaHandler
    from pydantic_core import CoreSchema
else:
    from pydantic.datetime_parse import parse_datetime


class DateTime(pendulum.DateTime):
    """https://stackoverflow.com/a/76719893"""

    @classmethod
    def __get_validators__(cls) -> Generator[Callable[[Any], pendulum.DateTime], None, None]:
        if not compat.IS_PYDANTIC_V2:
            yield parse_datetime
        yield cls._instance

    @classmethod
    def _instance(cls, v: datetime.datetime | pendulum.DateTime) -> pendulum.DateTime:
        return pendulum.instance(obj=v, tz=pendulum.UTC)

    if compat.IS_PYDANTIC_V2:

        @classmethod
        def __get_pydantic_core_schema__(
            cls, source_type: Any, handler: GetCoreSchemaHandler
        ) -> CoreSchema:
            return get_pydantic_core_schema(cls, handler(datetime.datetime))



from typing import Any

from pydantic import Field

from .. import container


class Query(container.BaseModel):
    def to_query(self) -> Any:
        ...


class OffsetPaginationQuery(Query):
    page: int | None = Field(None, ge=1)
    per_page: int | None = Field(None, ge=1)

    @property
    def can_paginate(self) -> bool:
        return self.page is not None and self.per_page is not None



import functools
from collections.abc import Callable
from typing import TypeVar

from pymongo.client_session import ClientSession, SessionOptions, TransactionOptions

from ..utils import unordered_partial

T = TypeVar("T")


class MongoUnitOfWork:
    def __init__(self, session_factory: Callable[..., ClientSession]):
        self.session_factory = session_factory

    def with_transaction(
        self,
        callback: Callable[[ClientSession], T],
        *,
        session_options: SessionOptions | None = None,
        transaction_options: TransactionOptions | None = None,
    ) -> T:
        session_options = session_options or SessionOptions()
        transaction_options = transaction_options or TransactionOptions()

        with self.session_factory(
            causal_consistency=session_options.causal_consistency,
            default_transaction_options=session_options.default_transaction_options,
            snapshot=session_options.snapshot,
        ) as session:
            result = session.with_transaction(
                callback=unordered_partial(p=functools.partial(callback), t=ClientSession),
                read_concern=transaction_options.read_concern,
                write_concern=transaction_options.write_concern,
                read_preference=transaction_options.read_preference,
                max_commit_time_ms=transaction_options.max_commit_time_ms,
            )

        return result



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from . import utils

if TYPE_CHECKING:
    import mypy_boto3_ssm


def get_parameters_from_ssm(
    ssm_client: mypy_boto3_ssm.SSMClient,
    prefix: str,
    preprocessors: dict[str, Callable[[str], str]] = None,
) -> dict[str, Any]:
    preprocessors = preprocessors or {}

    result = {}
    has_next = True
    next_token = None
    while has_next:
        params = {
            "Path": prefix,
            "WithDecryption": True,
            "Recursive": True,
        }
        if next_token is not None:
            params["NextToken"] = next_token

        response = ssm_client.get_parameters_by_path(**params)
        for parameter in response["Parameters"]:
            key = parameter["Name"].split("/")[-1].lower()
            value = parameter["Value"]
            result[key] = preprocessors.get(key, utils.ident)(value)

        next_token = response.get("NextToken")
        has_next = next_token is not None
    return result



import http
import secrets
from collections.abc import Callable
from typing import Annotated, Self

from fastapi import Depends, HTTPException
from fastapi.security import APIKeyHeader, HTTPBasic, HTTPBasicCredentials


class HTTPBasicAuth:
    def __init__(self, username: str, password: str, security: HTTPBasic):
        self.username = username
        self.password = password
        self.security = security

    def authenticator(self) -> Callable[[HTTPBasicCredentials], None]:
        security = self.security

        def compare_digest(a: str, b: str) -> bool:
            return secrets.compare_digest(a.encode(), b.encode())

        def authenticate(credentials: Annotated[HTTPBasicCredentials, Depends(security)]) -> None:
            is_correct_username = compare_digest(credentials.username, self.username)
            is_correct_password = compare_digest(credentials.password, self.password)

            if not (is_correct_username and is_correct_password):
                raise HTTPException(
                    status_code=http.HTTPStatus.UNAUTHORIZED,
                    detail="Incorrect username or password",
                    headers={"WWW-Authenticate": "Basic"},
                )

        return authenticate


class APIKeyAuth:
    """API Key Authentication for FastAPI

    Example:
    ```python
    import compose
    from fastapi import Depends
    from fastapi.security import APIKeyHeader

    app = FastAPI()
    api_key_header = APIKeyHeader("x-api-key", auto_error=False)

    # 고정 API Key 인증
    api_key_auth = compose.fastapi.APIKeyAuth.static(
        api_key="api-key",
        header=api_key_header,
    ).authenticator()

    # 동적 API Key 인증
    api_key_auth = compose.fastapi.APIKeyAuth(
        api_key_factory=api_key_factory,
        header=api_key_header,
    ).authenticator()

    @app.get("/auth/api-key", dependencies=[Depends(api_key_auth)])
    def authed_by_api_key():
        return {"message": "Authenticated"}
    """

    def __init__(self, api_key_factory: Callable[[], str], header: APIKeyHeader):
        self.api_key_factory = api_key_factory
        self.header = header

    def authenticator(self) -> Callable[[str], None]:
        _header = self.header

        def authenticate(header: Annotated[str, Depends(_header)]) -> None:
            api_key = self.api_key_factory()
            if header != api_key:
                raise HTTPException(
                    status_code=http.HTTPStatus.UNAUTHORIZED,
                    detail="Not authenticated. Invalid API key",
                )

        return authenticate

    @classmethod
    def static(cls, api_key: str, header: APIKeyHeader) -> Self:
        return cls(api_key_factory=lambda: api_key, header=header)



from .. import container


class AuthorizationGrant(container.BaseModel):
    access_token: str
    refresh_token: str | None = None


class UserResource(container.BaseModel):
    email: str
    name: str



try:
    import pytest  # noqa: F401
except ImportError:
    raise ImportError("Install `pytest` to use testing fixtures")

from .enums import *  # noqa: F401, F403
from .fixture import *  # noqa: F401, F403
from .hook import *  # noqa: F401, F403
from .param import *  # noqa: F401, F403



from collections.abc import Callable
from typing import Annotated, Any, TypeVar

from fastapi import Depends

from compose.container import BaseModel

U = TypeVar("U")
T = TypeVar("T", bound=BaseModel)


class CommandUpdater:
    def __init__(self, from_field: str, to_field: str):
        self.from_field = from_field
        self.to_field = to_field

    def __call__(self, cmd: T, user: U) -> T:
        return cmd.copy(update={self.to_field: getattr(user, self.from_field)}, deep=True)


class UserInjector:
    def __init__(
        self,
        user_getter: Callable[[], U],
        command_updater: Callable[[T, U], T],
    ):
        self.user_getter = user_getter
        self.command_updater = command_updater

    def injector(self) -> Callable[..., Any]:
        return self._injector

    def _injector(self, t: T, /) -> Callable[..., Any]:
        def inject_user(cmd: t, user: U = Depends(self.user_getter)) -> T:
            return self.command_updater(cmd, user)

        return inject_user


def create_with_user(
    user_injector: Callable[..., Any],
) -> Callable[[type[T]], type[T]]:
    def with_user(cmd: type[T]) -> type[T]:
        return Annotated[cmd, Depends(user_injector(cmd))]

    return with_user



import collections
import contextvars
from typing import TypeAlias

from .. import model
from .base import MessageQueue

EventMessageQueue: TypeAlias = collections.deque[model.EventMessage]

event_store: contextvars.ContextVar[EventMessageQueue] = contextvars.ContextVar("event_store")


class LocalMessageQueue(MessageQueue):
    def push(self, message: model.EventMessage) -> None:
        events = event_store.get()
        events.append(message)

    def peek(self) -> model.EventMessage | None:
        events = event_store.get()
        return events[0] if events else None

    def delete(self, message: model.EventMessage) -> None:
        events = event_store.get()

        if message in events:
            events.remove(message)



try:
    import opentelemetry  # noqa: F401
except ImportError:
    raise ImportError("Install `opentelemetry` extra to use opentelemetry features")

from .instrumentation.loguru.instrumentor import LoguruInstrumentor

__all__ = ["LoguruInstrumentor"]



from collections.abc import Callable
from typing import ClassVar

from authlib.integrations.httpx_client import AsyncOAuth2Client

from . import vo


class AuthorizationServer:
    headers: ClassVar[dict[str, str]] = {"Content-Type": "application/x-www-form-urlencoded"}
    token_path: ClassVar[str] = ""

    def __init__(self, base_url: str, auth_client_factory: Callable[..., AsyncOAuth2Client]):
        self.base_url = base_url
        self.auth_client_factory = auth_client_factory

    async def grant_authorization(self, redirect_uri: str, code: str) -> vo.AuthorizationGrant:
        async with self.auth_client_factory() as client:
            response = await client.fetch_token(
                url=f"{self.base_url}{self.token_path}",
                headers=self.headers,
                grant_type="authorization_code",
                client_id=client.client_id,
                client_secret=client.client_secret,
                redirect_uri=redirect_uri,
                code=code,
            )
        return vo.AuthorizationGrant.model_validate(response)

    async def renew_token(self, token: str) -> vo.AuthorizationGrant:
        async with self.auth_client_factory() as client:
            response = await client.fetch_token(
                url=self.token_path,
                headers=self.headers,
                grant_type="refresh_token",
                client_id=client.client_id,
                client_secret=client.client_secret,
                refresh_token=token,
            )
        return vo.AuthorizationGrant.model_validate(response)

    async def create_authorization_url(self, redirect_uri: str) -> str:
        raise NotImplementedError

    async def revoke_token(self, token: str) -> None:
        raise NotImplementedError



from collections.abc import Callable
from typing import Self

import pendulum
from authlib.jose import jwt

from .. import utils


class JWTIssuer:
    def __init__(
        self,
        secret_key: str,
        algorithm: str,
        issuer: str,
        token_id_generator: Callable[[], str],
        clock: type[pendulum.DateTime],
    ):
        self.secret_key = secret_key
        self.algorithm = algorithm
        self.issuer = issuer
        self.token_id_generator = token_id_generator
        self.clock = clock

    @classmethod
    def default(cls, secret_key: str, issuer: str) -> Self:
        return cls(
            secret_key=secret_key,
            algorithm="HS256",
            issuer=issuer,
            token_id_generator=utils.uuid4_hex,
            clock=pendulum.DateTime,
        )

    def issue(self, sub: str, expires_in: int, **kwargs) -> str:
        iat = self.clock.utcnow()
        return jwt.encode(
            header={"typ": "JWT", "alg": self.algorithm},
            payload={
                "sub": sub,
                "iss": self.issuer,
                "jti": self.token_id_generator(),
                "iat": int(iat.timestamp()),
                "exp": int(iat.add(seconds=expires_in).timestamp()),
                **kwargs,
            },
            key=self.secret_key,
        ).decode()



import types
from collections.abc import Callable
from typing import Any, TypeVar, cast

from compose import compat, typing

from .helper import CoreSchemaGettable

T = TypeVar("T")


def caster(factory: Callable[[Any], T], /) -> Callable[[Any], T]:
    def _cast(v: Any) -> T:
        return factory(v)

    return _cast


class Str(str, CoreSchemaGettable[str]):
    @classmethod
    def __get_validators__(cls) -> typing.ValidatorGenerator:
        if not compat.IS_PYDANTIC_V2:
            yield compat.str_validator
        yield caster(cls)


class Int(int, CoreSchemaGettable[int]):
    @classmethod
    def __get_validators__(cls) -> typing.ValidatorGenerator:
        if not compat.IS_PYDANTIC_V2:
            yield compat.int_validator
        yield caster(cls)


class Float(float, CoreSchemaGettable[float]):
    @classmethod
    def __get_validators__(cls) -> typing.ValidatorGenerator:
        if not compat.IS_PYDANTIC_V2:
            yield compat.float_validator
        yield caster(cls)


def _create_list_type(t: type[T], /) -> type[list[T]]:
    def __get_validators__(c) -> typing.ValidatorGenerator:
        if not compat.IS_PYDANTIC_V2:
            yield compat.list_validator
        yield caster(c)

    return cast(
        type[list[T]],
        types.new_class(
            f"{t.__name__.title()}List",
            (list[t], CoreSchemaGettable[list[t]]),
            exec_body=lambda ns: ns.update(
                {
                    "__get_validators__": classmethod(__get_validators__),
                }
            ),
        ),
    )


def create_list_type() -> Callable[[type[T]], type[list[T]]]:
    cache = {}

    def factory(t: type[T]) -> type[list[T]]:
        type_name = t.__name__

        if (cached := cache.get(type_name)) is not None:
            return cached

        _result = _create_list_type(t)
        cache[type_name] = _result
        return _result

    return factory


TypedList = create_list_type()
StrList = TypedList(str)
IntList = TypedList(int)



from __future__ import annotations

import enum
import inspect
from collections.abc import Iterable
from typing import Any, ClassVar, Generic, TypeVar, get_args, get_type_hints

import pendulum
import pymongo
from pymongo.client_session import ClientSession
from pymongo.collection import Collection
from pymongo.database import Database
from pymongo.results import UpdateResult

from .. import compat, types
from ..entity import Entity
from ..pagination import Pagination
from ..query.mongo import MongoFilterQuery, MongoQuery
from . import base

EntityType = TypeVar("EntityType", bound=Entity)
registry: dict[str, list[pymongo.IndexModel]] = {}


class SessionRequirement(str, enum.Enum):
    REQUIRED = "required"
    OPTIONAL = "optional"


def entity_to_mongo_schema(entity: EntityType, **kwargs) -> dict[str, Any]:
    default_kwargs = {"by_alias": True}
    return compat.model_dump(entity, **(default_kwargs | kwargs))


class MongoRepository(base.BaseRepository, Generic[EntityType]):
    """
    `MongoRepository` 추상 클래스

    상속 인자:
        `session_requirement`: `MongoRepository` 상속 시 `session_requirement` 인자를 통해
        `session` 인자 선언 여부를 검사할 수 있습니다. 프로덕션에서는 `session_requirement`를
        `SessionRequirement.REQUIRED`로 설정하는 것을 권장합니다.

        ```python
        class UserRepository(
            MongoRepository[User],
            session_requirement=SessionRequirement.REQUIRED,
        ):
            ...
        ```

    """

    __collection_name__: ClassVar[str] = ""
    __indexes__: ClassVar[list[pymongo.IndexModel] | None] = None

    def __init__(self, collection: Collection):
        self.collection = collection

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__()
        cls._validate_session_requirement(
            kwargs.get("session_requirement", SessionRequirement.OPTIONAL)
        )
        indexes = cls.__indexes__ or []
        if cls.__collection_name__ not in registry:
            registry[cls.__collection_name__] = indexes
            return

        index_documents = [idx.document for idx in registry[cls.__collection_name__]]
        registry[cls.__collection_name__].extend(
            [idx for idx in indexes if idx.document not in index_documents]
        )

    @classmethod
    def _validate_session_requirement(cls, requirement: SessionRequirement) -> None:
        if requirement == SessionRequirement.OPTIONAL:
            return

        error_message = "Method {} does not have a session argument"
        # 클래스 자체를 검사하기 때문에 `inspect.ismethod`를 사용하면 인스턴스 메서드가 아니라 클래스메서드를 리턴
        for name, member in inspect.getmembers(cls, predicate=inspect.isfunction):
            if name.startswith("__"):
                continue

            # https://stackoverflow.com/q/72510518
            # `inspect.get_annotations`를 사용하면 `__future__.annotations`을 사용했을 때에는 문자열을,
            # 그렇지 않은 경우에는 `type`을 리턴함. `__future__.annotations` 사용 여부에 관계 없이
            # 항상 `type`을 얻기 위해 `typing.get_type_hints` 사용
            if (type_hints := get_type_hints(member).get("session")) is None:
                raise ValueError(error_message.format(name))

            if not (args := get_args(type_hints)):
                raise ValueError(error_message.format(name))

            session = args[0]
            if session is not ClientSession:
                raise ValueError(error_message.format(name))

    @classmethod
    def create(cls, database: Database, **kwargs) -> MongoRepository:
        collection = database.get_collection(cls.__collection_name__, **kwargs)
        if (
            cls.__collection_name__ not in database.list_collection_names()
            and cls.__indexes__ is not None
        ):
            collection.create_indexes(cls.__indexes__)

        return cls(collection=collection)

    def find_by_id(
        self,
        entity_id: types.PyObjectId,
        session: ClientSession | None = None,
        **kwargs,
    ) -> EntityType | None:
        return self.find_by({"_id": entity_id}, session=session, **kwargs)

    def find_by(
        self, filter_: dict[str, Any], session: ClientSession | None = None, **kwargs
    ) -> EntityType | None:
        """https://stackoverflow.com/a/73746554/9331155"""
        entity_type: EntityType = get_args(self.__class__.__orig_bases__[0])[0]  # type: ignore
        result = self.collection.find_one(filter=filter_, session=session, **kwargs)
        return result and compat.model_validate(entity_type, result)

    def find_by_query(
        self, qry: MongoQuery, session: ClientSession | None = None, **kwargs
    ) -> dict[str, Any] | None:
        query_result = self.collection.aggregate(qry.to_query(), session=session, **kwargs)
        return next(query_result, None)

    def list_by_query(
        self, qry: MongoQuery, session: ClientSession | None = None, **kwargs
    ) -> list[dict[str, Any]]:
        query_result = self.collection.aggregate(qry.to_query(), session=session, **kwargs)
        return list(query_result)

    def add(self, entity: EntityType, session: ClientSession | None = None, **kwargs) -> None:
        self.collection.insert_one(entity_to_mongo_schema(entity), session=session, **kwargs)

    def add_many(
        self, entities: list[EntityType], session: ClientSession | None = None, **kwargs
    ) -> None:
        self.collection.insert_many(
            [entity_to_mongo_schema(entity) for entity in entities], session=session, **kwargs
        )

    def update(self, entity: EntityType, session: ClientSession | None = None, **kwargs) -> None:
        schema = entity_to_mongo_schema(entity)
        update_result = self.collection.update_one(
            {"_id": entity.id},
            {"$set": schema},
            session=session,
            **kwargs,
        )

        self.on_update(
            update_result=update_result,
            entity=entity,
            session=session,
            **kwargs,
        )

    def on_update(
        self,
        update_result: UpdateResult,
        entity: EntityType,
        session: ClientSession | None = None,
        **kwargs,
    ) -> None:
        if not update_result.modified_count:
            return

        if getattr(entity, "updated_at") is None:
            return

        self.collection.update_one(
            {"_id": entity.id},
            {"$set": {"updated_at": pendulum.DateTime.utcnow()}},
            session=session,
            **kwargs,
        )

    def update_many(
        self, entities: list[EntityType], session: ClientSession | None = None, **kwargs
    ) -> None:
        for entity in entities:
            self.update(entity=entity, session=session, **kwargs)

    def delete(
        self, entity_id: types.PyObjectId, session: ClientSession | None = None, **kwargs
    ) -> None:
        self.collection.delete_one({"_id": entity_id}, session=session, **kwargs)

    def execute_raw(
        self,
        operation: str,
        session: ClientSession | None = None,
        **operation_kwargs,
    ) -> Any:
        op = getattr(self.collection, operation, None)
        if op is None:
            raise ValueError(f"Unknown operation on collection: {operation}")
        return op(session=session, **operation_kwargs)

    def filter(
        self, qry: MongoFilterQuery, session: ClientSession | None = None, **kwargs
    ) -> Pagination:
        query_result = self.collection.aggregate(qry.to_query(), session=session, **kwargs)
        if (unwrapped := next(query_result, None)) is None:
            raise ValueError(f"{qry.__class__.__name__} returned nothing")

        return Pagination(
            total=(unwrapped["metadata"][0]["total"] if unwrapped["metadata"] else 0),
            items=unwrapped["items"],
            page=qry.page,
            per_page=qry.per_page,
        )


def setup_indexes(mongo_uri: str, db_names: Iterable[str]) -> None:
    # 동일한 이름의 컬렉션이 여러 DB에 존재할 경우, 컬렉션별로 인덱스를 생성할 수 없음
    with pymongo.MongoClient(mongo_uri) as client:
        for db_name in db_names:
            database = client.get_database(db_name)
            collection_names = database.list_collection_names()
            for collection_name, indexes in registry.items():
                if not collection_name or collection_name not in collection_names:
                    continue
                collection = database.get_collection(collection_name)

                previous_index_names = {index["name"] for index in collection.list_indexes()}
                current_index_names = {
                    *(index.document["name"] for index in indexes),
                    "_id_",
                }

                for index_name in previous_index_names - current_index_names:
                    collection.drop_index(index_name)

                if indexes:
                    collection.create_indexes(indexes)



from typing import Any, Generic, Self, TypeVar, get_args, get_origin

from pydantic import ValidationError

from .. import compat, container
from ..pagination import Pagination
from .extra import schema_by_field_name

if compat.IS_PYDANTIC_V2:
    from pydantic import ConfigDict
else:
    from pydantic.generics import GenericModel
    from pydantic.typing import is_union

ListItem = TypeVar("ListItem")


class Schema(container.BaseModel):
    if compat.IS_PYDANTIC_V2:
        model_config = ConfigDict(json_schema_extra=schema_by_field_name())
    else:

        class Config:
            schema_extra = schema_by_field_name()


class TimeStampedSchema(container.TimeStampedModel, Schema):
    ...


if compat.IS_PYDANTIC_V2:

    class ListSchema(Schema, Generic[ListItem]):
        total: int
        items: list[ListItem]

        @classmethod
        def from_pagination(
            cls,
            pagination: Pagination,
            parser_name: str = "model_validate",
            **parser_kwargs: Any,
        ) -> Self:
            if not pagination.items:
                return cls(**pagination.model_dump())

            annotation = cls.model_fields["items"].annotation
            item_type = get_args(annotation)[0]

            if not issubclass(item_type, container.BaseModel):
                data = pagination.model_dump(exclude={"extra"}) | pagination.extra
                return cls(**data)

            if (parser := getattr(item_type, parser_name, None)) is None:
                raise AttributeError(f"{item_type.__name__} has no attribute: {parser_name}")

            return cls(
                **pagination.model_dump(exclude={"items", "extra"}),
                **pagination.extra,
                items=[parser(item, **parser_kwargs) for item in pagination.items],
            )

else:

    class ListSchema(Schema, GenericModel, Generic[ListItem]):
        total: int
        items: list[ListItem]

        @classmethod
        def from_pagination(
            cls,
            pagination: Pagination,
            parser_name: str = "parse_obj",
            **parser_kwargs: Any,
        ) -> Self:
            if not pagination.items:
                return cls(**pagination.dict())

            item_type = cls.__fields__["items"].type_
            item_origin = get_origin(item_type)
            if is_union(item_origin):
                for arg in get_args(item_type):
                    try:
                        arg.parse_obj(pagination.items[0])
                        item_type = arg
                    except ValidationError:
                        continue

            item_parsable = issubclass(item_type, container.BaseModel)
            if not item_parsable:
                data = pagination.dict(exclude={"extra"}) | pagination.extra
                return cls(**data)

            if (parser := getattr(item_type, parser_name, None)) is None:
                raise AttributeError(f"{item_type.__name__} has no attribute: {parser_name}")

            return cls(
                **pagination.dict(exclude={"items", "extra"}),
                **pagination.extra,
                items=[parser(item, **parser_kwargs) for item in pagination.items],
            )


class InvalidParam(container.BaseModel):
    loc: str
    message: str
    type: str


class Error(container.BaseModel):
    title: str
    type: str
    detail: str | None = None
    invalid_params: list[InvalidParam] | None = None

    @classmethod
    def from_validation_error(cls, exc: ValidationError, title: str) -> Self:
        invalid_params = []
        for error in exc.errors():
            invalid_params.append(
                InvalidParam(
                    loc=".".join(str(v) for v in error["loc"]),
                    message=error["msg"],
                    type=error["type"],
                )
            )
        return cls(
            title=title,
            type="validation_error",
            invalid_params=invalid_params,
        )



from collections.abc import Callable, Generator
from typing import Any, Self, TypeAlias, TypeVar

from .types import PyObjectId

T = TypeVar("T")

Validator: TypeAlias = Callable[[Any], Self]
ValidatorGenerator: TypeAlias = Generator[Validator, None, None]

Factory: TypeAlias = Callable[..., T]
PyObjectIdFactory: TypeAlias = Factory[PyObjectId]

NoArgsFactory: TypeAlias = Callable[[], T]
NoArgsPyObjectIdFactory: TypeAlias = NoArgsFactory[PyObjectId]



from typing import Any, TypeVar

from pydantic import BaseModel

from . import utils

if utils.IS_PYDANTIC_V2:
    from pydantic import TypeAdapter
else:
    from pydantic import parse_obj_as

T = TypeVar("T")


def validate_obj(t: type[T], value: Any, /) -> T:
    if utils.IS_PYDANTIC_V2:
        return TypeAdapter(t).validate_python(value)
    else:
        return parse_obj_as(t, value)


def model_schema(t: type[BaseModel], **kwargs) -> dict[str, Any]:
    if utils.IS_PYDANTIC_V2:
        return t.model_json_schema(**kwargs)
    else:
        return t.schema(**kwargs)


def model_validate(t: type[BaseModel], obj: Any, **kwargs) -> BaseModel:
    if utils.IS_PYDANTIC_V2:
        return t.model_validate(obj, **kwargs)
    else:
        return t.parse_obj(obj, **kwargs)


def model_validate_json(t: type[BaseModel], obj: Any, **kwargs) -> BaseModel:
    if utils.IS_PYDANTIC_V2:
        return t.model_validate_json(obj, **kwargs)
    else:
        return t.parse_raw(obj, **kwargs)


def model_dump(t: BaseModel, **kwargs) -> dict[str, Any]:
    if utils.IS_PYDANTIC_V2:
        return t.model_dump(**kwargs)
    else:
        return t.dict(**kwargs)


def model_dump_json(t: BaseModel, **kwargs) -> str:
    if utils.IS_PYDANTIC_V2:
        return t.model_dump_json(**kwargs)
    else:
        return t.json(**kwargs)



from typing import Any

from .. import compat
from .schema import Schema

try:
    import orjson
except ImportError:
    raise ImportError("Install `orjson` to use orjson schema")


def orjson_dumps(v: Any, *, default):
    return orjson.dumps(v, default=default).decode()


class ORJSONSchema(Schema):
    if not compat.IS_PYDANTIC_V2:

        class Config:
            json_loads = orjson.loads
            json_dumps = orjson_dumps



from __future__ import annotations

import logging
import sys
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any

try:
    from loguru import logger
except ImportError:
    raise ImportError(
        "The `loguru` extra must be installed to use the `compose.logging` module. "
        "Install `compose` with `loguru` extra (`compose[loguru]`)"
    )

if TYPE_CHECKING:
    from loguru import Logger


class InterceptHandler(logging.Handler):
    """https://loguru.readthedocs.io/en/stable/overview.html#entirely-compatible-with-standard-logging"""  # noqa: E501

    def emit(self, record: logging.LogRecord) -> None:
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        frame, depth = logging.currentframe(), 0
        while frame and (depth == 0 or frame.f_code.co_filename == logging.__file__):
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())

    def filter(self, record: logging.LogRecord) -> bool:
        return "/health-check" not in record.getMessage()


def intercept_logging(intercept_handler: InterceptHandler, log_level: int) -> None:
    """Python 내장 logging 모듈을 loguru로 대체합니다. (해당 함수를 호출하려면 `loguru`를 설치해야 합니다.)"""
    logging.basicConfig(handlers=[intercept_handler], level=log_level, force=True)


def intercept(
    intercept_handler: InterceptHandler,
    log_names: Iterable[str] = (
        "gunicorn.error",
        "gunicorn.access",
        "uvicorn.error",
        "uvicorn.access",
    ),
) -> None:
    intercept_logging(intercept_handler=intercept_handler, log_level=logging.INFO)
    for name in log_names:
        logging.getLogger(name).handlers = [intercept_handler]


def get_default_logging_config(serialize_log: bool) -> dict[str, Any]:
    non_serialized_format = (
        "{time:YYYY-MM-DD HH:mm:ss.SSS} | <level>{level: <8}</level> | <level>{message}</level>"
    )
    return {
        "sink": sys.stdout,
        "diagnose": False,
        "format": "{message}" if serialize_log else non_serialized_format,
        "serialize": serialize_log,
    }


def get_default_logger(
    log_level: int,
    serialize_log: bool,
    **config: Any,
) -> Logger:
    intercept_handler = InterceptHandler()
    logging.basicConfig(handlers=[intercept_handler], level=log_level, force=True)
    intercept(intercept_handler=intercept_handler)
    logger.configure(
        handlers=[
            {
                **get_default_logging_config(serialize_log),
                **config,
            }
        ]
    )
    return logger



import asyncio
import logging

from .. import model
from ..messagebus import MessageBus
from ..queue.base import MessageQueue
from ..signal_handler import DefaultSignalHandler, SignalHandler
from .hook import DEFAULT_HOOKS, Hook, HookArgType, HookEventType, default_hook

logger = logging.getLogger("compose")


class MessageConsumer:
    def __init__(
        self,
        messagebus: MessageBus,
        message_queue: MessageQueue,
        hooks: dict[HookEventType, list[Hook]] | None = None,
        signal_handler: SignalHandler | None = None,
    ):
        self.messagebus = messagebus
        self.message_queue = message_queue
        self.hooks = DEFAULT_HOOKS | (hooks or {})
        self.signal_handler = signal_handler or DefaultSignalHandler()

        self._default_hook = default_hook

    async def run(self) -> None:
        self._execute_hook("on_start", "MessageConsumer started")

        while not self.signal_handler.received_signal:
            try:
                message = self.message_queue.peek()
            except Exception as exc:
                self._execute_hook("on_receive_error", exc)
                continue

            if message is None:
                continue
            self._execute_hook("on_receive", message)

            try:
                await asyncio.create_task(self.consume(message))
            except Exception as exc:
                self._execute_hook("on_consume_error", exc)
                continue
            self._execute_hook("on_consume", message)

    async def consume(self, message: model.EventMessage) -> None:
        await self.messagebus.handle_event(message.body)
        self.message_queue.delete(message)

    def _execute_hook(self, hook_event_type: HookEventType, arg: HookArgType, /) -> None:
        for hook in self.hooks.get(hook_event_type, [self._default_hook]):
            hook(arg)

    def shutdown(self) -> None:
        self._execute_hook("on_shutdown", "MessageConsumer shutting down")
        self.signal_handler.handle()



from __future__ import annotations

import copy
from typing import Any, ClassVar, Self, TypeVar

import inflection

DictExpression = dict[str, Any]
ListExpression = list[DictExpression]
Expression = TypeVar("Expression", DictExpression, ListExpression)

Id = "$_id"


class MongoKeyword(str):
    def __new__(cls, v: str):
        if v != _camelize(v):
            raise ValueError(f"Cannot interpret {v} as valid mongo keyword")

        return super().__new__(cls, v)

    @classmethod
    def from_py(cls, v: str) -> Self:
        return cls(_camelize(v))


def _camelize(v: str) -> str:
    return inflection.camelize(v.strip("_"), uppercase_first_letter=False)


class AggVar(str):
    """https://www.mongodb.com/docs/manual/reference/aggregation-variables/"""

    prefix: ClassVar[str] = "$"

    def __new__(cls, v: str | AggVar):
        copied = copy.deepcopy(v)
        num_prefixes = 0
        prefix_removed = not v.startswith(cls.prefix)
        while not prefix_removed:
            copied = copied[1:]
            num_prefixes += 1
            prefix_removed = not copied.startswith(cls.prefix)

        if num_prefixes not in (1, 2):
            raise ValueError(f"Cannot interpret {v} as valid aggregation variable")

        return super().__new__(cls, v)

    @classmethod
    def from_(cls, v: str) -> Self:
        if v.startswith(cls.prefix):
            raise ValueError(f"`v` must not start with `{cls.prefix}`")

        return cls(f"{cls.prefix * 2}{v}")

    @classmethod
    def current(cls, v: str) -> Self:
        if v.startswith(cls.prefix):
            raise ValueError(f"`v` must not start with `{cls.prefix}`")

        return cls(f"{cls.prefix}{v}")

    @classmethod
    def root(cls) -> Self:
        return cls.current("ROOT")


class _FieldPath(str):
    def __new__(cls, v: str):
        if not v.startswith("$"):
            raise ValueError("path must be prefixed with $")

        return super().__new__(cls, v)


class _String(str):
    def __new__(cls, v: str):
        if v.startswith("$"):
            raise ValueError("string must not be prefixed with $")

        return super().__new__(cls, v)



from collections.abc import Callable
from typing import Any, TypeVar, cast

from .base import GeneralAggregationOperator, Operator

OperatorType = TypeVar("OperatorType", bound=Operator)


def create_operator(
    name: str,
    expression_factory: Callable[[OperatorType], Any],
    base: tuple[type[OperatorType], ...],
) -> type[OperatorType]:
    return cast(
        type[OperatorType],
        type(name, base, {"expression": expression_factory}),
    )


def create_general_aggregation_operator(
    name: str, mongo_operator: str
) -> type[GeneralAggregationOperator]:
    return cast(
        type[GeneralAggregationOperator],
        type(name, (GeneralAggregationOperator,), {"mongo_operator": mongo_operator}),
    )



from typing import Any, Self

from ..base import Evaluable, Operator
from ..types import DictExpression


class IfNull(Operator):
    def __init__(
        self,
        input_expr: str,
        non_null_input_expr: Any,
        replacement_expr: Any | None = None,
        /,
    ):
        """
        Aggregation operator `$ifNull`

        Reference:
            https://www.mongodb.com/docs/manual/reference/operator/aggregation/ifNull/#mongodb-expression-exp.-ifNull
        """

        self.input_expr = input_expr
        self.non_null_input_expr = non_null_input_expr
        self.replacement_expr = replacement_expr

    @classmethod
    def single(cls, input_expr: str, non_null_input_expr: Any, /) -> Self:
        return cls(input_expr, non_null_input_expr)

    @classmethod
    def multiple(cls, input_expr: str, non_null_input_expr: Any, replacement_expr: Any, /) -> Self:
        return cls(input_expr, non_null_input_expr, replacement_expr)

    def expression(self) -> DictExpression:
        expressions = [self.input_expr, self.non_null_input_expr]
        if self.replacement_expr is not None:
            expressions.append(self.replacement_expr)

        return {"$ifNull": [Evaluable(expr).expression() for expr in expressions]}


class Cond(Operator):
    def __init__(self, if_: Any, then: Any, else_: Any):
        """
        Aggregation operator `$cond`

        Reference:
            https://www.mongodb.com/docs/manual/reference/operator/aggregation/cond/#mongodb-expression-exp.-cond
        """

        self.if_ = if_
        self.then = then
        self.else_ = else_

    def expression(self) -> DictExpression:
        return {
            "$cond": {
                "if": Evaluable(self.if_).expression(),
                "then": Evaluable(self.then).expression(),
                "else": Evaluable(self.else_).expression(),
            }
        }



import inspect
from collections.abc import Callable
from typing import Any, TypeVar

from fastapi import Depends

from compose.dependency.wiring import Provider

F = TypeVar("F", bound=Callable[..., Any])


def auto_wired(provider: Provider) -> Callable[[F], F]:
    """
    FastAPI 엔드포인트에 의존성을 자동으로 주입하는 데코레이터. 해당 데코레이터는 `@inject` 데코레이터보다
    먼저 적용되어야 합니다.

    주의: `__signature__` 속성은 일반 함수에만 존재하므로 해당 데코레이터는 메서드에 사용할 수 없습니다.
    """

    def wrapper(f: F) -> F:
        signature = inspect.signature(f)

        updated_params = []
        for name, param in signature.parameters.items():
            updated_param = param
            try:
                provided = provider(param.annotation)
                updated_param = updated_param.replace(default=Depends(provided))
            except ValueError:
                pass

            updated_params.append(updated_param)

        f.__signature__ = signature.replace(parameters=updated_params)

        return f

    return wrapper



from typing import TypeVar

from compose import container
from compose.event import Event

MessageBody = TypeVar("MessageBody", bound=Event)


class EventMessage(container.BaseModel):
    body: MessageBody


class SqsEventMessage(EventMessage):
    receipt_handle: str



from __future__ import annotations

from collections.abc import Callable, Collection
from typing import TYPE_CHECKING, Any

import wrapt
from loguru import logger
from opentelemetry import trace
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.utils import unwrap

ORIGINAL_PATCHER_ATTR = "_otel_original_patcher"
_instruments = tuple()

if TYPE_CHECKING:
    from loguru import Logger


def trace_injector(
    tracer_provider: trace.TracerProvider,
) -> Callable[[dict[str, Any]], dict[str, Any]]:
    def inject_trace(record: dict[str, Any]) -> dict[str, Any]:
        span = trace.get_current_span()
        ctx = span.get_span_context()
        is_valid_span = span != trace.INVALID_SPAN

        resource = getattr(tracer_provider, "resource", None)
        trace_record = {
            "otel_service_name": (
                resource.attributes.get("service.name", "") if resource is not None else ""
            ),
            "otel_trace_id": (
                trace.format_trace_id(ctx.trace_id)
                if is_valid_span
                else str(trace.INVALID_TRACE_ID)
            ),
            "otel_span_id": (
                trace.format_span_id(ctx.span_id) if is_valid_span else str(trace.INVALID_SPAN_ID)
            ),
            "otel_trace_sampled": ctx.trace_flags.sampled if is_valid_span else False,
        }

        extra = record.get("extra", {})
        return {"extra": extra | trace_record}

    return inject_trace


def create_record_patcher(
    inject_trace: Callable[[dict[str, Any]], dict[str, Any]],
) -> Callable[[dict[str, Any]], None]:
    def patcher(record: dict[str, Any]) -> None:
        record |= inject_trace(record)

    return patcher


def create_configure_wrapper(
    inject_trace: Callable[[dict[str, Any]], dict[str, Any]],
) -> Callable[..., list[int]]:
    def wrapped_configure(
        func: Callable[..., list[int]],
        instance: Logger,
        *args,
        **kwargs,
    ) -> list[int]:
        """https://github.com/DataDog/dd-trace-py/blob/main/ddtrace/contrib/loguru/patch.py"""

        original_patcher = kwargs.get("patcher")
        setattr(instance, ORIGINAL_PATCHER_ATTR, original_patcher)

        if original_patcher is None:
            return func(*args, **kwargs)

        def wrapped_patcher(record: dict[str, Any]) -> None:
            original_patcher(record)
            record |= inject_trace(record)

        kwargs["patcher"] = wrapped_patcher
        return func(*args, **kwargs)

    return wrapped_configure


def default_record_patcher(_: dict[str, Any]) -> None:
    return None


class LoguruInstrumentor(BaseInstrumentor):
    """

    Reference:
        https://github.com/DataDog/dd-trace-py/tree/main/ddtrace/contrib/loguru
        https://github.com/open-telemetry/opentelemetry-python-contrib/tree/main/instrumentation/opentelemetry-instrumentation-logging
    """

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs) -> None:
        tracer_provider = kwargs.get("tracer_provider", trace.get_tracer_provider())

        inject_trace = trace_injector(tracer_provider)
        logger.configure(patcher=create_record_patcher(inject_trace))
        wrapt.wrap_function_wrapper(
            logger,
            "configure",
            create_configure_wrapper(inject_trace),
        )

    def _uninstrument(self, **kwargs) -> None:
        original_patcher = getattr(logger, ORIGINAL_PATCHER_ATTR, None)
        has_original_patcher = original_patcher is not None
        if has_original_patcher:
            logger.configure(patcher=original_patcher)

        unwrap(logger, "configure")
        if not has_original_patcher:
            logger.configure(patcher=default_record_patcher)



import enum
import functools
from collections.abc import Awaitable, Callable
from typing import Any, Self, TypeAlias

from fastapi import Request, Response

from compose import container

from .exception_handler import ExceptionHandler

try:
    import sentry_sdk
    from sentry_sdk.integrations import Integration
except ImportError:
    raise ImportError("Install `sentry-sdk` to use sentry helpers")


class Level(enum.StrEnum):
    INFO = enum.auto()
    WARNING = enum.auto()
    ERROR = enum.auto()


class ErrorEvent(container.BaseModel):
    level: Level

    @classmethod
    def for_info(cls) -> Self:
        return cls(level=Level.INFO)

    @classmethod
    def for_warning(cls) -> Self:
        return cls(level=Level.WARNING)

    @classmethod
    def for_error(cls) -> Self:
        return cls(level=Level.ERROR)


SentryHook: TypeAlias = Callable[[dict[str, Any], dict[str, Any]], dict[str, Any]]


def create_before_send_hook(
    error: dict[str, ErrorEvent], default_error_level: Level = Level.WARNING
) -> SentryHook:
    def before_send(event: dict[str, Any], hint: dict[str, Any]) -> dict[str, Any]:
        exc_name = ""
        if "exc_info" in hint:
            exc_type, *_ = hint["exc_info"]
            exc_name = exc_type.__name__

        error_event = error.get(exc_name)
        event["level"] = default_error_level if error_event is None else error_event.level
        return event

    return before_send


def init_sentry(
    integrations: list[Integration],
    environment: str,
    tags: dict[str, str],
    dsn: str | None = None,
    before_send: SentryHook | None = None,
    **kwargs,
) -> None:
    if dsn is None:
        return

    sentry_sdk.init(
        dsn=dsn,
        integrations=integrations,
        environment=environment,
        before_send=before_send,
        **kwargs,
    )
    scope = sentry_sdk.Scope.get_current_scope()
    for key, value in tags.items():
        scope.set_tag(key, value)


def capture_error(handler: ExceptionHandler) -> ExceptionHandler:
    @functools.wraps(handler)
    async def wrapper(request: Request, exc: Exception) -> Response | Awaitable[Response]:
        sentry_sdk.capture_exception(exc)
        result = handler(request, exc)
        if isinstance(result, Awaitable):
            return await result
        return result

    return wrapper



import pendulum
from authlib.jose import JWTClaims, errors, jwt

from .. import exceptions


class JWTDecoder:
    def __init__(self, secret_key: str, clock: type[pendulum.DateTime]):
        self.secret_key = secret_key
        self.clock = clock

    def decode(self, token: str) -> JWTClaims:
        try:
            decoded = jwt.decode(token.encode(), key=self.secret_key)
        except (errors.DecodeError, errors.BadSignatureError):
            raise exceptions.AuthorizationError("Invalid token")

        try:
            decoded.validate(now=int(self.clock.utcnow().timestamp()))
        except (errors.ExpiredTokenError, errors.InvalidClaimError):
            raise exceptions.AuthorizationError("Expired or incorrect format")

        return decoded



import urllib.parse
from typing import Self


class ContentDisposition(str):
    @classmethod
    def attachment(cls, filename: str) -> Self:
        return cls(f"attachment; filename*=UTF-8''{urllib.parse.quote(filename)}")



from .authorization_server import AuthorizationServer
from .resource_server import ResourceServer
from .token_decoder import JWTDecoder
from .token_issuer import JWTIssuer
from .vo import AuthorizationGrant, UserResource

__all__ = [
    "AuthorizationServer",
    "AuthorizationGrant",
    "ResourceServer",
    "UserResource",
    "JWTDecoder",
    "JWTIssuer",
]

try:
    from .password import HashedPassword  # noqa: F401

    __all__.append("HashedPassword")
except ImportError:
    pass



from typing import Any

from ..base import Evaluable, Operator
from ..types import DictExpression


class SetIntersection(Operator):
    def __init__(self, *values: Any):
        self.values = list(values)

    def expression(self) -> DictExpression:
        return Evaluable({"$setIntersection": self.values}).expression()



import functools
import time
import types
from typing import ClassVar, NewType, Protocol, Self

import pendulum
import pymongo
from pymongo.collection import Collection
from pymongo.database import Database
from pymongo.errors import PyMongoError

from .exceptions import LockAcquisitionFailedError

Seconds = NewType("Seconds", float)


class MongoLockAcquirer(Protocol):
    def __call__(
        self,
        key: str,
        auto_release_after: Seconds = Seconds(60),
        timeout: Seconds = Seconds(60),
        lock_acquisition_interval: Seconds = Seconds(0.1),
    ) -> Self: ...


class MongoLock:
    default_collection_name: ClassVar[str] = "compose.lock"
    index_name: ClassVar[str] = "expires_at_ttl"

    def __init__(
        self,
        key: str,
        collection: Collection,
        auto_release_after: Seconds = Seconds(60),
        timeout: Seconds = Seconds(60),
        lock_acquisition_interval: Seconds = Seconds(0.1),
    ):
        self.key = key
        self.collection = collection
        self.auto_release_after = auto_release_after
        self.timeout = timeout
        self.lock_acquisition_interval = lock_acquisition_interval

    def _ensure_ttl_index(self) -> None:
        current_index = self.collection.index_information()
        if self.index_name not in current_index:
            self.collection.create_index(
                [(self.index_name, pymongo.ASCENDING)],
                name=self.index_name,
                expireAfterSeconds=0,
            )

    @classmethod
    def acquirer(cls, db: Database, collection_name: str | None = None) -> MongoLockAcquirer:
        collection_name = collection_name or cls.default_collection_name
        return functools.partial(cls, collection=db[collection_name])

    def __enter__(self) -> Self:
        if self.acquire():
            return self

        raise LockAcquisitionFailedError(f"Failed to acquire lock for key: {self.key}")

    def __exit__(
        self,
        exc_type: type[BaseException],
        exc_val: BaseException,
        exc_tb: types.TracebackType,
    ) -> None:
        self.release()

    def acquire(self) -> bool:
        start_time = pendulum.DateTime.utcnow()
        try_until = start_time.add(seconds=self.timeout)
        expires_at = pendulum.DateTime.utcnow().add(seconds=self.auto_release_after)

        while True:
            try:
                lock = self.collection.find_one_and_update(
                    {
                        "_id": self.key,
                        "expires_at": {"$lt": pendulum.DateTime.utcnow()},
                    },
                    {"$set": {"expires_at": expires_at}},
                    upsert=True,
                    return_document=pymongo.ReturnDocument.AFTER,
                )
                if lock is None:
                    continue

                # MongoDB는 Date 타입에서 나노초를 버리기 때문에 <=로 비교
                if pendulum.instance(lock["expires_at"], tz=pendulum.UTC) <= expires_at:
                    return True

            except PyMongoError:
                pass

            if pendulum.DateTime.utcnow() > try_until:
                return False

            time.sleep(self.lock_acquisition_interval)

    def release(self) -> None:
        self.collection.delete_one({"_id": self.key})



import enum
from typing import Any, ClassVar, Self, cast


class BaseError(Exception):
    default_message: ClassVar[str | None] = None

    def __init__(
        self,
        message: str | enum.Enum | None = None,
        detail: Any | None = None,
        invalid_params: list[dict[str, Any]] | None = None,
    ):
        if isinstance(message, enum.Enum):
            message = message.value
        super().__init__(message)

        if message is None and self.default_message is None:
            raise ValueError("`message` or `default_message` must not be None")

        self.message = str(message) if message is not None else self.default_message
        self.detail = detail
        self.invalid_params = invalid_params

    def __str__(self):
        return self.message

    @classmethod
    def with_default_message(cls, name: str, message: str) -> type[Self]:
        return cast(
            type[Self],
            type(name, (cls,), {"default_message": message}),
        )


AuthorizationError = BaseError.with_default_message("AuthorizationError", "Authorization failed")
NotAllowedError = BaseError.with_default_message("NotAllowedError", "Not allowed")
DoesNotExistError = BaseError.with_default_message("DoesNotExistError", "Resource not found")
DomainValidationError = BaseError.with_default_message("DomainValidationError", "Validation failed")



from .container import DeclarativeContainer
from .wiring import (
    DEFAULT_RESOLVABLE_PROVIDER_TYPES,
    ConflictResolution,
    create_provider,
    create_wirer,
    provide,
    resolve,
    resolve_by_object_name,
)

__all__ = [
    "create_wirer",
    "resolve",
    "provide",
    "ConflictResolution",
    "resolve_by_object_name",
    "create_provider",
    "DeclarativeContainer",
    "DEFAULT_RESOLVABLE_PROVIDER_TYPES",
]



import enum


def parametrize_enum_flag_property_test(
    e: type[enum.Enum],
    truthy: list[enum.Enum] | tuple[enum.Enum] | set[enum.Enum],
) -> list[tuple[enum.Enum, bool]]:
    return [(value, value in truthy) for value in e]



from .. import compat
from .datetime import DateTime
from .helper import CoreSchemaGettable, SupportsGetValidators, chain
from .object_id import PyObjectId
from .url import S3ContentUrl
from .vo import Float, Int, IntList, Str, StrList, TypedList
from .web import ContentDisposition

__all__ = [
    "PyObjectId",
    "DateTime",
    "SupportsGetValidators",
    "chain",
    "CoreSchemaGettable",
    "Float",
    "Int",
    "Str",
    "IntList",
    "StrList",
    "TypedList",
    "S3ContentUrl",
    "ContentDisposition",
]

if compat.IS_PYDANTIC_V2:
    from .helper import get_pydantic_core_schema  # noqa: F401

    __all__.extend(["get_pydantic_core_schema"])



import abc
from typing import Any

from .. import base


class MongoQuery(base.Query, abc.ABC):
    @abc.abstractmethod
    def to_query(self) -> list[dict[str, Any]]:
        raise NotImplementedError


class MongoFilterQuery(base.OffsetPaginationQuery, MongoQuery):
    @abc.abstractmethod
    def to_query(self) -> list[dict[str, Any]]:
        raise NotImplementedError


MongoOffsetFilterQuery = MongoFilterQuery



from __future__ import annotations

import json
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, Self, TypeAlias, TypeVar

import bson
from pydantic import BaseModel as PydanticBaseModel

from . import compat, field, types

if compat.IS_PYDANTIC_V2:
    from pydantic import ConfigDict

    AbstractSetIntStr: TypeAlias = set[int] | set[str]
    MappingIntStrAny: TypeAlias = dict[int, Any] | dict[str, Any]
else:
    if TYPE_CHECKING:
        from pydantic.typing import AbstractSetIntStr, DictStrAny, MappingIntStrAny

IncEx: TypeAlias = set[int] | set[str] | dict[int, Any] | dict[str, Any] | None
Model = TypeVar("Model", bound=PydanticBaseModel)


class BaseModel(PydanticBaseModel):
    if compat.IS_PYDANTIC_V2:

        def json(
            self,
            indent: int | None = None,
            include: IncEx = None,
            exclude: IncEx = None,
            by_alias: bool = False,
            exclude_unset: bool = False,
            exclude_defaults: bool = False,
            exclude_none: bool = False,
            round_trip: bool = False,
            warnings: bool = True,
        ) -> str:
            return self.model_dump_json(
                indent=indent,
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                round_trip=round_trip,
                warnings=warnings,
                serialize_as_any=True,
            )

        def dict(
            self,
            include: IncEx = None,
            exclude: IncEx = None,
            by_alias: bool = False,
            exclude_unset: bool = False,
            exclude_defaults: bool = False,
            exclude_none: bool = False,
            round_trip: bool = False,
            warnings: bool = True,
        ):
            return self.model_dump(
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                round_trip=round_trip,
                warnings=warnings,
                serialize_as_any=True,
            )

        def copy(
            self,
            *,
            include: AbstractSetIntStr | MappingIntStrAny | None = None,
            exclude: AbstractSetIntStr | MappingIntStrAny | None = None,
            update: dict[str, Any] | None = None,
            deep: bool = False,
        ) -> Model:
            return super().model_copy(update=update, deep=deep)

        def encode(
            self,
            *,
            indent: int | None = None,
            include: IncEx = None,
            exclude: IncEx = None,
            by_alias: bool = False,
            exclude_unset: bool = False,
            exclude_defaults: bool = False,
            exclude_none: bool = False,
            round_trip: bool = False,
            warnings: bool = True,
        ) -> dict[str, Any]:
            return json.loads(
                self.model_dump_json(
                    indent=indent,
                    include=include,
                    exclude=exclude,
                    by_alias=by_alias,
                    exclude_unset=exclude_unset,
                    exclude_defaults=exclude_defaults,
                    exclude_none=exclude_none,
                    round_trip=round_trip,
                    warnings=warnings,
                )
            )

        # TODO[pydantic]: The following keys were removed: `json_encoders`.
        # Check https://docs.pydantic.dev/dev-v2/migration/#changes-to-config for more information.
        model_config = ConfigDict(
            populate_by_name=True,
            validate_assignment=True,
            extra="ignore",
        )

    else:

        @classmethod
        def model_validate(cls, obj: Any, **kwargs: Any) -> Self:
            return cls.parse_obj(obj)

        def copy(
            self,
            *,
            include: AbstractSetIntStr | MappingIntStrAny | None = None,
            exclude: AbstractSetIntStr | MappingIntStrAny | None = None,
            update: DictStrAny | None = None,
            deep: bool = False,
            strict: bool = True,
        ) -> Self:
            if strict and (diff := set((update or {}).keys()) - set(self.__fields__.keys())):
                raise AttributeError(
                    f"{self.__class__.__name__} has no attributes: {', '.join(diff)}"
                )

            return super().copy(
                include=include,
                exclude=exclude,
                update=update,
                deep=deep,
            )

        def encode(
            self,
            *,
            include: AbstractSetIntStr | MappingIntStrAny | None = None,
            exclude: AbstractSetIntStr | MappingIntStrAny | None = None,
            by_alias: bool = False,
            skip_defaults: bool | None = None,
            exclude_unset: bool = False,
            exclude_defaults: bool = False,
            exclude_none: bool = False,
            encoder: Callable[[Any], Any] | None = None,
            models_as_dict: bool = True,
            **dumps_kwargs: Any,
        ) -> dict[str, Any]:
            return self.__config__.json_loads(
                self.json(
                    include=include,
                    exclude=exclude,
                    by_alias=by_alias,
                    skip_defaults=skip_defaults,
                    exclude_unset=exclude_unset,
                    exclude_defaults=exclude_defaults,
                    exclude_none=exclude_none,
                    encoder=encoder,
                    models_as_dict=models_as_dict,
                    **dumps_kwargs,
                )
            )

        class Config:
            json_encoders = {bson.ObjectId: str}
            allow_population_by_field_name = True
            validate_assignment = True


class TimeStampedModel(BaseModel):
    created_at: types.DateTime = field.DateTimeField()
    updated_at: types.DateTime = field.DateTimeField()

    if not compat.IS_PYDANTIC_V2:

        def __init_subclass__(cls, **kwargs: Any) -> None:
            super().__init_subclass__(**kwargs)

            created_at_field = cls.__fields__.pop("created_at")
            updated_at_field = cls.__fields__.pop("updated_at")
            cls.__fields__ |= dict(created_at=created_at_field, updated_at=updated_at_field)



import signal


class SignalHandler:
    def __init__(self):
        self._received_signal = False

        signal.signal(signal.SIGINT, self._handle_signal)
        signal.signal(signal.SIGTERM, self._handle_signal)

    def _handle_signal(self, signum, frame):
        self._received_signal = True

    @property
    def received_signal(self) -> bool:
        return self._received_signal



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Self

import compose
from compose import utils

from .. import model
from .base import MessageQueue

if TYPE_CHECKING:
    import mypy_boto3_sqs

try:
    import boto3  # noqa: F401
except ImportError:
    raise ImportError(
        "The `aws` extra must be installed to use the `SqsMessageQueue`. "
        "Install compose with `aws` extra (`compose[aws]`)"
    )


class SqsMessageQueue(MessageQueue):
    def __init__(
        self,
        sqs_client: mypy_boto3_sqs.SQSClient,
        queue_name: str,
        wait_time_seconds: int,
        event_cls_map: dict[str, type[compose.event.Event]],
    ):
        self.client = sqs_client
        self.wait_time_seconds = wait_time_seconds
        self.event_cls_map = event_cls_map

        self._queue_url = self.client.get_queue_url(QueueName=queue_name)["QueueUrl"]

    @classmethod
    def with_event_cls(
        cls,
        sqs_client: mypy_boto3_sqs.SQSClient,
        queue_name: str,
        wait_time_seconds: int,
        event_cls: type[compose.event.Event],
    ) -> Self:
        return cls(
            sqs_client=sqs_client,
            queue_name=queue_name,
            wait_time_seconds=wait_time_seconds,
            event_cls_map={
                event_cls.__name__: event_cls for event_cls in utils.descendants_of(event_cls)
            },
        )

    def push(self, message: model.SqsEventMessage) -> None:
        self.client.send_message(
            QueueUrl=self._queue_url,
            MessageBody=json.dumps(message.body.encode(by_alias=True)),
            MessageGroupId=str(message.body.id),
            MessageDeduplicationId=str(message.body.id),
            MessageAttributes=dict(
                event_name=dict(
                    DataType="String",
                    StringValue=message.body.__class__.__name__,
                )
            ),
        )

    def peek(self) -> model.SqsEventMessage | None:
        response = self.client.receive_message(
            QueueUrl=self._queue_url,
            AttributeNames=["All"],
            MessageAttributeNames=["All"],
            MaxNumberOfMessages=1,
            WaitTimeSeconds=self.wait_time_seconds,
        )

        messages = response.get("Messages", [])
        if not messages:
            return None

        message = next(iter(messages))
        event_name = message["MessageAttributes"]["event_name"]["StringValue"]
        event_cls = self.event_cls_map[event_name]
        event_body = json.loads(message["Body"])

        return model.SqsEventMessage(
            body=event_cls(**event_body),
            receipt_handle=message["ReceiptHandle"],
        )

    def delete(self, message: model.SqsEventMessage) -> None:
        self.client.delete_message(QueueUrl=self._queue_url, ReceiptHandle=message.receipt_handle)



import abc
import logging
import threading

logger = logging.getLogger("compose")


class SignalHandler(abc.ABC):
    @abc.abstractmethod
    def handle(self) -> None:
        raise NotImplementedError

    @property
    def received_signal(self) -> bool:
        raise NotImplementedError


class DefaultSignalHandler(SignalHandler):
    def __init__(self):
        self._received_signal = False

    def handle(self) -> None:
        self._received_signal = True

    @property
    def received_signal(self) -> bool:
        return self._received_signal


class ThreadSignalHandler(SignalHandler):
    def __init__(self):
        self._exit_event = threading.Event()

    def handle(self) -> None:
        self._exit_event.set()

    @property
    def received_signal(self) -> bool:
        return self._exit_event.is_set()
